@@ -27,6 +27,7 @@ using namespace std;
 #define LOCATE_DB "/var/lib/mlocate/mlocate.db"
 #define UPDATEDB "updatedb.mlocate"
 #else
+#include "nolocate.h"
 #define LOCATE_DB "/var/lib/plocate/plocate.db"
 #define UPDATEDB "updatedb.plocate"
 #endif
@@ -98,11 +99,17 @@ static bool updatedb()
 
 static void one_file(const string& path)
 {
+	struct stat buffer;
+	if (stat(path.c_str(), &buffer)) {
+		cerr << "file not found\n";
+		exit(1);
+	}
+
 	// is it a static file ?
-	dpkg_start();
+	dpkg_start("/");
 	if (query(path.c_str())) {
 		dpkg_end();
-		return;
+		exit(0);
 	}
 
 	string infile = path;
@@ -115,8 +122,7 @@ static void one_file(const string& path)
 	char* file = realpath(infile.c_str(), nullptr);
 
 	vector<string> packages;
-	vector<string> dpkg;
-	read_dpkg(packages, dpkg, false);
+	read_dpkg_header(packages);
 	dpkg_end();
 
 	// is it a dynamic file ?
@@ -125,7 +131,7 @@ static void one_file(const string& path)
 	for (const auto& gl: globs) {
 		if (myglob(file, gl.path)) {
 			cout << gl.package << '\n';
-			return;
+			exit(0);
 		};
 	}
 
@@ -135,11 +141,12 @@ static void one_file(const string& path)
 	for (const auto& ex: explain) {
 		if (infile == ex.path) {
 			cout << ex.package << '\n';
-			return;
+			exit(0);
                 }
         }
 
 	cerr << "no matching package found\n";
+	exit(1);
 }
 
 static int one_package(const string& package)
@@ -173,6 +180,7 @@ static const char* const default_filter_dir = "/etc/cruft/filters/";
 static const char* const default_ignore_file = "/etc/cruft/ignore";
 static const char* const default_ruleset_file = "/usr/share/cruft/ruleset";
 static const char* const default_bugs_file = "/usr/share/cruft/bugs";
+static const char* const default_root_dir = "/";
 
 static void print_help_message()
 {
@@ -187,6 +195,10 @@ static void print_help_message()
 	cout << "    -I --ignore      path for ignore file (default: " << default_ignore_file << ")\n";
 	cout << "    -R --ruleset     path for ruleset file (default: " << default_ruleset_file << ")\n";
 	cout << "    -B --bugs        path for known bugs file (default: " << default_bugs_file << ")\n";
+#ifndef BUSTER
+	cout << "    -l --locate      do use locate (deprecated)\n";
+	cout << "    -r --root        root directory (default: " << default_root_dir << ", does not work with --locate)\n";
+#endif
 
 	cout << '\n';
 
@@ -197,7 +209,9 @@ static void cruft(const string& ignore_file,
                   const string& filter_dir,
                   const string& ruleset_file,
                   const string& explain_dir,
-                  const string& bugs_file)
+                  const string& bugs_file,
+                  bool locate,
+                  const string& root_dir)
 {
 	bool debug = getenv("DEBUG") != nullptr;
 
@@ -211,19 +225,25 @@ static void cruft(const string& ignore_file,
 	strftime(buf, sizeof(buf), "%c", timeinfo);
 	cout << "cruft report: " << buf << '\n';
 
-	bool updated = updatedb();
-	if (!updated) {
-		cerr << "warning: plocate database is outdated" << endl << flush;
+	if (locate) {
+		bool updated = updatedb();
+		if (!updated) {
+			cerr << "warning: plocate database is outdated" << endl << flush;
+		}
+		elapsed("updatedb");
 	}
-	elapsed("updatedb");
 
 	vector<string> fs;
-	thread thr_plocate(read_locate, ref(fs), ignore_file);
+#ifndef BUSTER
+	thread thr_plocate(locate ? read_locate : read_nolocate, ref(fs), ignore_file, root_dir);
+#else
+	thread thr_plocate(read_locate, ref(fs), ignore_file, root_dir);
+#endif
 
 	vector<string> packages;
 	vector<string> dpkg;
-	dpkg_start();
-	thread thr_dpkg(read_dpkg, ref(packages), ref(dpkg), false);
+	dpkg_start(root_dir);
+	thread thr_dpkg(read_dpkg, ref(packages), ref(dpkg), false, root_dir);
 	thr_plocate.join();
 	thr_dpkg.join();
 	dpkg_end();
@@ -288,17 +308,27 @@ static void cruft(const string& ignore_file,
 	read_filters(filter_dir, ruleset_file, packages, globs);
 	elapsed("read filters");
 	vector<string> cruft3;
+	vector<bool> used_globs(globs.size(), false);
 	for (const auto& cr: cruft) {
 		bool match=false;
-		for (const auto& gl: globs) {
-			match=myglob(cr, gl.path);
-			if (match) break;
+		for (size_t i = 0; i < globs.size(); i++) {
+			match=myglob(cr, globs[i].path);
+			if (match) {
+				used_globs[i] = true;
+				break;
+			}
 		}
 		if (!match) cruft3.push_back(cr);
 	}
+	if (debug)
+		for (size_t i = 0; i < globs.size(); i++)
+			if (!used_globs[i])
+				cout << "unused ruleset: " << globs[i].package << " " << globs[i].path << endl;
 	elapsed("extra vs globs");
 	if (debug) cerr << cruft3.size() << " files in cruft3 database\n\n";
 
+	// set CRUFT_ROOT for explain scripts
+	setenv("CRUFT_ROOT", root_dir == "/" ? "" : root_dir.c_str(), 0);
 	// match the dynamic "explain" filters
 	vector<owner> explain;
 	read_explain(explain_dir, packages, explain);
@@ -341,27 +371,32 @@ static void cruft(const string& ignore_file,
 int main(int argc, char *argv[])
 {
 	bool do_one_package = false;
+	bool locate = false;
 	string package = "";
 	string explain_dir = default_explain_dir;
 	string filter_dir = default_filter_dir;
 	string ignore_file = default_ignore_file;
 	string ruleset_file = default_ruleset_file;
 	string bugs_file = default_bugs_file;
+	string root_dir = default_root_dir;
 
 	const struct option long_options[] =
 	{
 		{"help", no_argument, nullptr, 'h'},
+		{"locate", no_argument, nullptr, 'l'},
+		{"no-locate", no_argument, nullptr, 'n'},  // no longer used
 		{"package", required_argument, nullptr, 'p'},
 		{"explain", required_argument, nullptr, 'E'},
 		{"filter", required_argument, nullptr, 'F'},
 		{"ignore", required_argument, nullptr, 'I'},
 		{"ruleset", required_argument, nullptr, 'R'},
 		{"bugs", required_argument, nullptr, 'B'},
+		{"root", required_argument, nullptr, 'r'},
 		{0, 0, 0, 0}
 	};
 
 	int opt, opti = 0;
-	while ((opt = getopt_long(argc, argv, "p:E:F:hI:R:B:", long_options, &opti)) != 0) {
+	while ((opt = getopt_long(argc, argv, "p:E:F:hI:nR:B:r:", long_options, &opti)) != 0) {
 		if (opt == EOF)
 			break;
 
@@ -390,6 +425,10 @@ int main(int argc, char *argv[])
 			ignore_file = optarg;
 			break;
 
+		case 'l':
+			locate = true;
+			break;
+
 		case 'R':
 			ruleset_file = optarg;
 			break;
@@ -402,6 +441,12 @@ int main(int argc, char *argv[])
 			print_help_message();
 			exit(1);
 
+		case 'r':
+			root_dir = optarg;
+			if (!root_dir.empty() && root_dir.back() != '/')
+				root_dir += '/';
+			break;
+
 	        default:
 	            cerr << "Invalid getopt return value: " << opt << "\n";
 				break;
@@ -412,16 +457,8 @@ int main(int argc, char *argv[])
 	if (do_one_package) exit(one_package(package));
 
 	if (optind < argc) {
-		if (optind + 1 == argc) {
-			struct stat buffer;
-			if (stat(argv[1], &buffer) == 0) {
-				one_file(argv[1]);
-				exit(0);
-			} else {
-				cerr << "file not found\n";
-				exit(1);
-			}
-		}
+		if (optind + 1 == argc)
+			one_file(argv[1]);
 
 		cerr << "Invalid non-option arguments:";
 		while (optind < argc)
@@ -432,5 +469,5 @@ int main(int argc, char *argv[])
 	}
 
 	// else: standard cruft report
-	cruft(ignore_file, filter_dir, ruleset_file, explain_dir, bugs_file);
+	cruft(ignore_file, filter_dir, ruleset_file, explain_dir, bugs_file, locate, root_dir);
 }
